/**
 * P4P Vuze Plugin
 * 
 * Copyright 2008, Yale Laboratory of Networked Systems
 */

package net.yale.cs.p4p.p2p.azureus.client.data;

import java.util.ArrayList;
import java.util.Collection;

import net.yale.cs.p4p.p2p.azureus.client.P4PPlugin;

public abstract class QueryCache
{
	protected abstract class QueryTask extends Thread {

		private int interval_;
		protected void setInterval(int value) {
			interval_ = value;
		}

		public QueryTask(int interval) {
			interval_ = interval;
		}

		protected abstract void query() throws InterruptedException;
		
		public void run() {
			try {
				while (!interrupted()) {
					query();
					sleep(interval_);
				}
			} catch (InterruptedException e) {
				/* Let thread exit */
			}
		}
	}

	private P4PPlugin plugin_;
	public P4PPlugin getPlugin() { return plugin_; }

	private int numThreads_;
	private Collection<Thread> queryThreads_ = new ArrayList<Thread>();

	protected abstract QueryTask makeQueryTask();

	public QueryCache(P4PPlugin plugin, int numThreads) {
		plugin_ = plugin;
		numThreads_ = numThreads;
	}

	public void start() {
		for (int i = 0; i < numThreads_; ++i)
			queryThreads_.add(makeQueryTask());
		for (Thread queryThread : queryThreads_)
			queryThread.start();
	}

	public void interrupt() {
		for (Thread queryThread : queryThreads_)
			queryThread.interrupt();
	}

	public void join() throws InterruptedException {
		for (Thread queryThread : queryThreads_)
			queryThread.join();
	}
	
	public void joinNoThrow() {
		while (true) {
			try {
				join();
				break;
			} catch (InterruptedException e) {
				/* Try again */
			}
		}
	}

}
