package net.yale.cs.p4p.p2p.azureus.client.data;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import net.yale.cs.p4p.p2p.azureus.client.P4PPlugin;
import net.yale.cs.p4p.p2p.azureus.client.data.selection.PeerSelectionBase;
import net.yale.cs.p4p.p2p.azureus.client.data.selection.PeerSelectionBase.ScheduledPeer;
import net.yale.cs.p4p.p2p.azureus.client.data.selection.PeerSelectionBase.ScheduledPeerInfo;
import net.yale.cs.p4p.p2p.azureus.client.hacks.DiscoveredPeersHack;

import org.gudy.azureus2.plugins.download.DownloadStats;
import org.gudy.azureus2.plugins.peers.Peer;
import org.gudy.azureus2.plugins.peers.PeerDescriptor;
import org.gudy.azureus2.plugins.torrent.Torrent;

public class TorrentState
{
	private static final boolean DEBUG_LOG = false;
	
	private static final PeerCollection EMPTY_PEER_COLLECTION = new PeerCollection();
	
	public static final String peerDescriptorToString(PeerDescriptor peer) {
		return new StringBuffer()
			.append(peer.getIP())
			.append(':').append(peer.getTCPPort())
			.append(':').append(peer.getUDPPort())
			.append(':').append(DiscoveredPeersHack.getPeerHandshakeType(peer))
			.toString();
	}

	public static final class PeerRecord {

		private PeerDescriptor peer_;
		public PeerDescriptor getPeer() { return peer_; }

		public String getPeerSource() { return DiscoveredPeersHack.getPeerSource(peer_); }

		private long lastSeen_;
		public long getLastSeen() { return lastSeen_; }
		private void updateLastSeen() {
			lastSeen_ = System.currentTimeMillis();
		}
		
		private ScheduledPeerInfo schedInfo_;
		public ScheduledPeerInfo getSchedInfo() { return schedInfo_; }

		private PeerRecord(PeerDescriptor peer, ScheduledPeerInfo schedInfo, long lastSeen) {
			peer_ = peer;
			schedInfo_ = schedInfo;
			lastSeen_ = lastSeen;
		}

		private PeerRecord(PeerDescriptor peer, ScheduledPeerInfo schedInfo) {
			this(peer, schedInfo, System.currentTimeMillis());
		}

		private PeerRecord(PeerRecord peer, ScheduledPeerInfo schedInfo) {
			this(peer.getPeer(), schedInfo, peer.getLastSeen());
		}

		private PeerRecord(PeerDescriptor peer) {
			this(peer, PeerSelectionBase.EMPTY_SCHEDULED_PEER_INFO);
		}

		public void updateIfRequired(PeerDescriptor peer) {
			if (DiscoveredPeersHack.getPeerHandshakeType(peer) > DiscoveredPeersHack.getPeerHandshakeType(peer_))
				peer_ = peer;
		}
		
		private boolean equalsPeerDescriptor(PeerDescriptor peer) {
			return peer_.getIP().equals(peer.getIP())
				&& peer_.getTCPPort() == peer.getTCPPort()
				&& peer_.getUDPPort() == peer.getUDPPort();
		}

		private boolean equalsPeer(Peer peer) {
			return peer_.getIP().equals(peer.getIp())
				&& peer_.getTCPPort() == peer.getTCPListenPort()
				&& peer_.getUDPPort() == peer.getUDPListenPort();
		}
		
		public boolean equals(Object o) {
			if (!(o instanceof PeerRecord))
				return false;
			return equalsPeerDescriptor(((PeerRecord)o).peer_);
		}

		public int hashCode() {
			return peer_.hashCode();
		}
		
	}

	private static final class PeerCollection {

		private static PeerRecord getPeerInBucket(Set<PeerRecord> peers, PeerDescriptor peer) {
			for (PeerRecord p : peers)
				if (p.equalsPeerDescriptor(peer))
					return p;
			return null;
		}

		private static PeerRecord getPeerInBucket(Set<PeerRecord> peers, Peer peer) {
			for (PeerRecord p : peers)
				if (p.equalsPeer(peer))
					return p;
			return null;
		}

		private LinkedHashSet<PeerRecord> orderedPeers_ = new LinkedHashSet<PeerRecord>();
		private Map<String, Set<PeerRecord>> peers_ = new HashMap<String, Set<PeerRecord>>();
		
		public PeerRecord getRecord(Peer peer) {
			Set<PeerRecord> bucket = getBucket(peer.getIp(), false);
			if (bucket == null)
				return null;
			return getPeerInBucket(bucket, peer);
		}

		public LinkedHashSet<PeerRecord> getAll() {
			return new LinkedHashSet<PeerRecord>(orderedPeers_);
		}

		public List<PeerDescriptor> getAllDesc() {
			List<PeerDescriptor> result = new ArrayList<PeerDescriptor>(orderedPeers_.size());
			for (PeerRecord peer : orderedPeers_)
				result.add(peer.getPeer());
			return result;
		}
		
		public int size() {
			return orderedPeers_.size();
		}

		private Set<PeerRecord> getBucket(String ip, boolean autoCreate) {
			Set<PeerRecord> result = peers_.get(ip);
			if (result == null && autoCreate)
				peers_.put(ip, result = new HashSet<PeerRecord>());
			return result;
		}

		public PeerRecord remove(PeerRecord peer) {
			Set<PeerRecord> bucket = getBucket(peer.getPeer().getIP(), false);
			if (bucket == null)
				return null;
			if (!bucket.remove(peer))
				return null;
			orderedPeers_.remove(peer);
			return peer;
		}

		public PeerRecord remove(PeerDescriptor peer) {
			Set<PeerRecord> bucket = getBucket(peer.getIP(), false);
			if (bucket == null)
				return null;
			PeerRecord rec = getPeerInBucket(bucket, peer);
			if (rec == null)
				return null;
			if (!bucket.remove(rec))
				return null;
			orderedPeers_.remove(rec);
			return rec;
		}
		
		public PeerRecord remove(Peer peer) {
			Set<PeerRecord> bucket = getBucket(peer.getIp(), false);
			if (bucket == null)
				return null;
			PeerRecord rec = getPeerInBucket(bucket, peer);
			if (rec == null)
				return null;
			if (!bucket.remove(rec))
				return null;
			orderedPeers_.remove(rec);
			return rec;
		}

		public boolean add(PeerDescriptor peer, boolean doUpdate) {
			Set<PeerRecord> bucket = getBucket(peer.getIP(), true);
			PeerRecord existingPeer = getPeerInBucket(bucket, peer);
			if (existingPeer == null) {
				PeerRecord rec = new PeerRecord(peer);
				bucket.add(rec);
				orderedPeers_.add(rec);
				return true;
			} else {
				existingPeer.updateIfRequired(peer);
				if (doUpdate)
					existingPeer.updateLastSeen();
				return false;
			}
		}

		public boolean add(PeerRecord peer) {
			boolean result = getBucket(peer.getPeer().getIP(), true).add(peer);
			if (result)
				orderedPeers_.add(peer);
			return result;
		}

		public boolean contains(PeerDescriptor peer) {
			Set<PeerRecord> bucket = getBucket(peer.getIP(), false);
			if (bucket == null)
				return false;
			PeerRecord rec = getPeerInBucket(bucket, peer);
			if (rec != null)
				rec.updateIfRequired(peer);
			return rec != null;
		}

		public boolean contains(Peer peer) {
			Set<PeerRecord> bucket = getBucket(peer.getIp(), false);
			if (bucket == null)
				return false;
			return getPeerInBucket(bucket, peer) != null;
		}
	}
	
	public static final class PeerSnapshot {
		private List<PeerDescriptor> availPeers_;
		public List<PeerDescriptor> getAvailPeers() { return availPeers_; }
		
		private List<PeerDescriptor> queuedPeers_;
		public List<PeerDescriptor> getQueuedPeers() { return queuedPeers_; }
		
		private List<PeerDescriptor> connPeers_;
		public List<PeerDescriptor> getConnectedPeers() { return connPeers_; }

		public int size() {
			return availPeers_.size() + queuedPeers_.size() + connPeers_.size();
		}
		
		public boolean isEmpty() {
			return size() == 0;
		}

		private PeerSnapshot(PeerCollection availPeers, PeerCollection queuedPeers, PeerCollection connPeers) {
			availPeers_ = availPeers.getAllDesc();
			queuedPeers_ = queuedPeers.getAllDesc();
			connPeers_ = connPeers.getAllDesc();
		}
	}

	private P4PPlugin plugin_;
	private Torrent torrent_;
	public Torrent getTorrent() { return torrent_; }

	/* All peers that we know about for this torrent */
	private PeerCollection allPeers_ = new PeerCollection();

	/* Peers that have not yet been queued or connected to */
	private PeerCollection availPeers_ = new PeerCollection();
	
	/* Peers that are currently queued to be connected with */
	private PeerCollection queuedPeers_ = new PeerCollection();

	/* Peers that are currently connected */
	private PeerCollection connPeers_ = new PeerCollection();
	
	/* True if torrent contains Vuze-biased peers */
	private boolean isVuzeBiased_ = false;
	public synchronized boolean getVuzeBiased() { return isVuzeBiased_; }
	public synchronized void setVuzeBiased() { isVuzeBiased_ = true; }

	public TorrentState(P4PPlugin plugin, Torrent torrent) {
		plugin_ = plugin;
		torrent_ = torrent;
	}

	private void logStates(StringBuffer sb, PeerCollection peers) {
		sb.append("\tCount: ").append(peers.size()).append('\n');
		for (PeerRecord peer : peers.getAll())
			sb.append('\t').append(peerDescriptorToString(peer.getPeer())).append('\n');
	}

	private void logStates(String action) {
		StringBuffer sb = new StringBuffer();
		sb.append("Torrent: ").append(torrent_.getName()).append('\n');
		sb.append("Action: ").append(action).append('\n');
		sb.append("All:\n");
		logStates(sb, allPeers_);
		sb.append("Connected:\n");
		logStates(sb, connPeers_);
		sb.append("Queued:\n");
		logStates(sb, queuedPeers_);
		sb.append("Available:\n");
		logStates(sb, availPeers_);

		System.out.println(sb.toString());
	}

	private void checkInvariant(String action) {
		if (allPeers_.size() != connPeers_.size() + queuedPeers_.size() + availPeers_.size()) {
			logStates(action);
			throw new RuntimeException("Invalid state: Invariant violated at action " + action);
		}
	}

	private void debugAction(String action) {
		checkInvariant(action);

		if (DEBUG_LOG)
			logStates(action);
	}
	
	/* Find a particular peer in one of our states */
	private PeerRecord removeExistingPeer(Peer peer) {
		PeerRecord rec = availPeers_.remove(peer);
		if (rec != null)
			return rec;
		rec = queuedPeers_.remove(peer);
		if (rec != null)
			return rec;
		rec = connPeers_.remove(peer);
		if (rec != null)
			return rec;
		throw new RuntimeException("Illegal state: Failed to find existing peer " + peer.getIp());
	}

	/* Find a particular peer in one of our states */
	private PeerRecord removeExistingPeer(PeerDescriptor peer) {
		PeerRecord rec = availPeers_.remove(peer);
		if (rec != null)
			return rec;
		rec = queuedPeers_.remove(peer);
		if (rec != null)
			return rec;
		rec = connPeers_.remove(peer);
		if (rec != null)
			return rec;
		throw new RuntimeException("Illegal state: Failed to find existing peer " + peer.getIP());
	}

	/* Find a particular peer in one of our states */
	private PeerRecord removeExistingPeer(PeerRecord peer) {
		PeerRecord rec = availPeers_.remove(peer);
		if (rec != null)
			return rec;
		rec = queuedPeers_.remove(peer);
		if (rec != null)
			return rec;
		rec = connPeers_.remove(peer);
		if (rec != null)
			return rec;
		throw new RuntimeException("Illegal state: Failed to find existing peer " + peer.getPeer().getIP());
	}

	/* Peer newly discovered from a peer source; place in 'available' state */
	private synchronized void peerDiscovered(PeerDescriptor peer, boolean doUpdate) {
		debugAction("ENTER peerDiscovered");
		try {
			/* Simply update timestamp if we already know about it. This doesn't
			 * change the peer's current state. */
			if (!allPeers_.add(peer, doUpdate))
				return;

			/* We didn't already know about it; initial state is 'available' */
			availPeers_.add(peer, false);
		} finally {
			debugAction("EXIT peerDiscovered");
		}
	}

	/* Peers newly discovered from a peer source; place in 'available' state */
	public synchronized void peersDiscovered(Collection<PeerDescriptor> peers, boolean doUpdate) {
		debugAction("ENTER peersDiscovered");
		try {
			for (PeerDescriptor peer : peers)
				peerDiscovered(peer, doUpdate);
		} finally {
			debugAction("EXIT peersDiscovered");
		}
	}

	/* Select peer for scheduling; move from 'available' to 'queued' state */
	private synchronized boolean peerQueued(ScheduledPeer peer) {
		debugAction("ENTER peerQueued");
		try {
			/* Ignore if we don't know about it. */
			if (!allPeers_.contains(peer.getPeer()))
				return false;

			/* If already connected, ignore it */
			if (connPeers_.contains(peer.getPeer()))
				return false;

			/* The scheduling information will be available
			 * in the returned PeerRecord on the connection event */
			return queuedPeers_.add(new PeerRecord(removeExistingPeer(peer.getPeer()), peer.getScheduleInfo()));
		} finally {
			debugAction("EXIT peerQueued");
		}
	}

	/* Select peers for scheduling; move from 'available' to 'queued' state.
	 * NOTE: Input list of peers is modified to only contain the peers that
	 *       were actually enqueued, and then the modified list is returned.
	 *       See comments in peerQueued(peer) for more information.
	 */
	public synchronized List<ScheduledPeer> peersQueued(List<ScheduledPeer> peers) {
		debugAction("ENTER peersQueued");
		try {
			for (Iterator<ScheduledPeer> itr = peers.iterator(); itr.hasNext(); )
				if (!peerQueued(itr.next()))
					itr.remove();
			return peers;
		} finally {
			debugAction("EXIT peersQueued");
		}
	}

	/* Get set of peers currently in 'queued' state */
	public synchronized List<PeerDescriptor> getQueuedPeerDesc() {
		debugAction("ENTER getQueuedPeerDesc");
		try {
			return queuedPeers_.getAllDesc();
		} finally {
			debugAction("EXIT getQueuedPeerDesc");
		}
	}

	/* Move all 'queued' peers to 'available' state and then return all 'available' peers;
	 * also cleans out peers that have aged out */
	public synchronized PeerSnapshot getSchedulablePeers(DownloadStats stats, int curQueued) {
		debugAction("ENTER getSchedulablePeers");
		try {
			/* Special case: if the current set of 'discovered' peers maintained by Vuze
			 * is empty, but we think there should be something in there, it means that
			 * we never received the peerConnected() state change event.  It appears that
			 * we are never notified of all peers in all cases. In this case, we will 
			 * simply remove our current set of queued peers so we will avoid rescheduling
			 * them. */
			if (curQueued == 0) {
				for (PeerRecord peer : queuedPeers_.getAll()) {
					queuedPeers_.remove(peer);
					allPeers_.remove(peer);
				}
			}

			/* Move 'queued' peers to 'available' state */
			for (PeerRecord peer : queuedPeers_.getAll())
				availPeers_.add(queuedPeers_.remove(peer));

			/* Compute expected lifetime of a peer in the swarm based on our
			 * own current download speed */
			long dlSpeedBps = stats != null ? stats.getDownloadAverage() : 0;
			dlSpeedBps = Math.min(plugin_.getConfig().getMaxDownloadBPS(), dlSpeedBps);
			dlSpeedBps = Math.max(plugin_.getConfig().getMinDownloadBPS(), dlSpeedBps);
			long expectedLifetimeMs = (torrent_.getSize() / dlSpeedBps) * 1000L;
			
			/* Compute time before which we consider a its 'last-seen' time too old */
			long oldTime = System.currentTimeMillis() - (plugin_.getConfig().getPeerLifetimeMultiplier() * expectedLifetimeMs);

			/* Remove the old peers from our state */
			for (PeerRecord peer : allPeers_.getAll()) {
				if (peer.getLastSeen() < oldTime) {
					allPeers_.remove(peer);
					removeExistingPeer(peer);
				}
			}
			
			/* Get all 'available' and 'disconnected' peers */
			return new PeerSnapshot(availPeers_, EMPTY_PEER_COLLECTION, EMPTY_PEER_COLLECTION);
		} finally {
			debugAction("EXIT getSchedulablePeers");
		}
	}

	/* Move peer from either 'available' or 'queued' to 'connected' state and return its full information */
	public synchronized PeerRecord peerConnected(Peer peer) {
		debugAction("ENTER peerConnected");
		try {
			/* Ignore if we don't know about it. For example, it may
			 * have been an incoming peer. */
			if (!allPeers_.contains(peer))
				return null;
	
			/* It is possible to get connected events for peers we didn't
			 * explicitly queue. In this case, set the peer to 'connected'
			 * regardless of its current state.
			 */
			PeerRecord rec = removeExistingPeer(peer);
			connPeers_.add(rec);
	
			return rec;
		} finally {
			debugAction("EXIT peerConnected");
		}
	}

	/* Move peer from 'connected' to 'disconnected' state */
	public synchronized void peerDisconnected(Peer peer) {
		debugAction("ENTER peerDisconnected");
		try {

			/* Ignore if we never knew anything about it */
			if (!allPeers_.contains(peer))
				return;
	
			/* It is possible to get disconnected events for peers we didn't
			 * know were connected. In this case, we'll just remove the peer
			 * and mark it as 'disconnected' regardless of what state we think
			 * it is in to avoid reconnecting to it immediately. */
			removeExistingPeer(peer);
			allPeers_.remove(peer);
		} finally {
			debugAction("EXIT peerDisconnected");
		}
	}

	public synchronized PeerSnapshot getSnapshot() {
		debugAction("ENTER getSnapshot");
		try {
			return new PeerSnapshot(availPeers_, queuedPeers_, connPeers_);
		} finally {
			debugAction("EXIT getSnapshot");
		}
	}
	
}
