package net.yale.cs.p4p.p2p.azureus.client.data.selection;

import net.yale.cs.p4p.p2p.azureus.client.P4PPlugin;
import net.yale.cs.p4p.p2p.azureus.client.hacks.DiscoveredPeersHack;

import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.peers.Peer;
import org.gudy.azureus2.plugins.peers.PeerManager;
import org.gudy.azureus2.plugins.torrent.Torrent;

public class PeerConnectionMonitor {

	private P4PPlugin plugin_;
	private PeerSelectionBase peerSelection_;

	public PeerConnectionMonitor(P4PPlugin plugin, PeerSelectionBase peerSelection) {
		plugin_ = plugin;
		peerSelection_ = peerSelection;
	}

	public synchronized void updateOptimisticUnchoke(Download download) {
		Torrent torrent = download.getTorrent();
		if (torrent == null)
			return;
		
		Peer peer = peerSelection_.selectOptimisticUnchoke(download);
		if (peer == null)
			return;
		
		plugin_.getLogger().log(torrent, "Selected peer " + peer.getIp() + " to unchoke");
		
		peer.setOptimisticUnchoke(true);
	}

	public synchronized void reschedule(Download download, boolean force) {
		/* Select more peers if forced or needed */
		if (force || needsPeers(download))
			peerSelection_.updatePeers(download);
	}

	/* Check to see if the download needs more peers */
	private boolean needsPeers(Download download) {
		PeerManager peerManager = download.getPeerManager();
		if (peerManager == null)
			return false;

		Torrent torrent = download.getTorrent();
		if (torrent == null)
			return false;

		int count = DiscoveredPeersHack.getDiscoveredPeersCount(peerManager);
		plugin_.getLogger().log(torrent, "Discovered peers queue has " + count + " peers");

		return count == 0;
	}
	
}
