package net.yale.cs.p4p.p2p.azureus.client.ui;

import java.util.Collection;
import java.util.HashSet;
import java.util.Locale;

import net.yale.cs.p4p.api.data.PID;
import net.yale.cs.p4p.p2p.azureus.client.P4PPlugin;
import net.yale.cs.p4p.p2p.azureus.client.data.PeerCache;
import net.yale.cs.p4p.p2p.azureus.client.data.TorrentState;
import net.yale.cs.p4p.p2p.azureus.client.data.Torrents;
import net.yale.cs.p4p.p2p.azureus.client.data.TorrentState.PeerSnapshot;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.List;
import org.gudy.azureus2.plugins.peers.Peer;
import org.gudy.azureus2.plugins.peers.PeerDescriptor;
import org.gudy.azureus2.plugins.ui.UIInstance;
import org.gudy.azureus2.plugins.ui.UIManager;
import org.gudy.azureus2.plugins.ui.UIManagerListener;
import org.gudy.azureus2.plugins.ui.tables.TableCell;
import org.gudy.azureus2.plugins.ui.tables.TableCellRefreshListener;
import org.gudy.azureus2.plugins.ui.tables.TableColumn;
import org.gudy.azureus2.plugins.ui.tables.TableManager;
import org.gudy.azureus2.ui.swt.plugins.UISWTInstance;
import org.gudy.azureus2.ui.swt.plugins.UISWTView;
import org.gudy.azureus2.ui.swt.plugins.UISWTViewEvent;
import org.gudy.azureus2.ui.swt.plugins.UISWTViewEventListener;

public class MainView {

	private static final String PID_COLUMN_ID = "PID";

	private static final String VIEW_ID = "P4P";

	private class UIAttachListener implements UIManagerListener {
		public void UIAttached(UIInstance instance) {
			if (!(instance instanceof UISWTInstance))
				return;
			addView((UISWTInstance)instance);
		}
		public void UIDetached(UIInstance instance) {
			if (!(instance instanceof UISWTInstance))
				return;
			removeView((UISWTInstance)instance);
		}
	}

	private final class ViewListener implements UISWTViewEventListener {

		private UISWTInstance instance_;
		public UISWTInstance getInstance() { return instance_; }

		public ViewListener(UISWTInstance instance) {
			instance_ = instance;
		}

		private List torrentsList_;
		private List availPeersList_;
		private List queuedPeersList_;
		private List connPeersList_;

		public boolean eventOccurred(UISWTViewEvent event) {
			switch (event.getType()) {
				case UISWTViewEvent.TYPE_CREATE:
					/* FIXME: Return 'false' if view already opened to prevent
					 * 	  opening multiple copies of the view */
					event.getView().setControlType(UISWTView.CONTROLTYPE_SWT);
					return true;
				case UISWTViewEvent.TYPE_INITIALIZE:
					doInitialize(event.getView(), (Composite)event.getData());
					return true;
				case UISWTViewEvent.TYPE_REFRESH:
					doRefresh(event.getView());
					return true;
				case UISWTViewEvent.TYPE_LANGUAGEUPDATE:
					doChangeLanguage(event.getView(), Locale.getDefault());
					break;
			}
			return true;
		}

		private void doInitialize(UISWTView view, Composite parent) {
			GridLayout layout = new GridLayout(5, false);
			parent.setLayout(layout);

			torrentsList_ = new List(parent, SWT.BORDER | SWT.SINGLE | SWT.V_SCROLL);
			torrentsList_.setLayoutData(new GridData(GridData.FILL_BOTH));
			
			availPeersList_ = new List(parent, SWT.BORDER | SWT.SINGLE | SWT.V_SCROLL);
			availPeersList_.setLayoutData(new GridData(GridData.FILL_BOTH));
			
			queuedPeersList_ = new List(parent, SWT.BORDER | SWT.SINGLE | SWT.V_SCROLL);
			queuedPeersList_.setLayoutData(new GridData(GridData.FILL_BOTH));
			
			connPeersList_ = new List(parent, SWT.BORDER | SWT.SINGLE | SWT.V_SCROLL);
			connPeersList_.setLayoutData(new GridData(GridData.FILL_BOTH));

			torrentsList_.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent e) {
					doTorrentSelected();
				}
			});
		}

		private String getCurTorrent() {
			int selected = torrentsList_.getSelectionIndex();
			return (selected >= 0) ? torrentsList_.getItem(selected) : null;
		}

		private void clearTorrentData() {
			availPeersList_.removeAll();
			queuedPeersList_.removeAll();
			connPeersList_.removeAll();
		}

		private void fillList(List list, Collection<PeerDescriptor> peers) {
			list.removeAll();
			for (PeerDescriptor peer : peers) {
				StringBuffer sb = new StringBuffer();
				sb.append(peer.getIP()).append(':').append(peer.getTCPPort());
				PID pid = peerCache_.getPID(peer);
				if (pid != null)
					sb.append(" (").append(pid).append(')');
				list.add(sb.toString());
			}
		}

		private void doTorrentSelected() {
			
			String curTorrent = getCurTorrent();
			if (curTorrent == null) {
				clearTorrentData();
				return;
			}
			
			TorrentState state = torrents_.getTorrentByName(curTorrent);
			if (state == null) {
				clearTorrentData();
				return;
			}

			PeerSnapshot peers = state.getSnapshot();
			fillList(availPeersList_, peers.getAvailPeers());
			fillList(queuedPeersList_, peers.getQueuedPeers());
			fillList(connPeersList_, peers.getConnectedPeers());
		}

		private void doRefresh(UISWTView view) {

			String prevTorrent = getCurTorrent();
			torrentsList_.removeAll();
			int selIdx = -1;
			int curIdx = 0;
			for (TorrentState torrent : torrents_.getAll()) {
				if (prevTorrent != null && prevTorrent.equals(torrent.getTorrent().getName()))
					selIdx = curIdx;
				torrentsList_.add(torrent.getTorrent().getName(), curIdx++);
			}
			if (selIdx >= 0)
				torrentsList_.select(selIdx);
			
			doTorrentSelected();
		}

		private void doChangeLanguage(UISWTView view, Locale locale) {
			
		}
		
	}

	private final class PIDColumnRefreshListener implements TableCellRefreshListener {

		public void refresh(TableCell cell) {
			Peer peer = (Peer)cell.getDataSource();
			if (peer == null)
				return;
			
			cell.setText(peerCache_.lookupPID(peer.getIp()).toString());
		}
		
	}
	
	private P4PPlugin plugin_;
	private Torrents torrents_;
	private PeerCache peerCache_;
	private UIManager uiManager_;
	private TableManager tableManager_;
	
	private TableColumn pidColumn_;
	private PIDColumnRefreshListener pidColumnRefreshListener_;

	private UIAttachListener uiAttachListener_ = new UIAttachListener();
	private Collection<UISWTInstance> openUIs_ = new HashSet<UISWTInstance>();

	public MainView(P4PPlugin plugin) {
		plugin_ = plugin;
		torrents_ = plugin_.getTorrents();
		peerCache_ = plugin_.getPeerCache();
		uiManager_ = plugin_.getPluginInterface().getUIManager();
		uiManager_.addUIListener(uiAttachListener_);
		tableManager_ = uiManager_.getTableManager();
		addPIDColumn();
	}
	
	private synchronized void addPIDColumn() {
		pidColumn_ = tableManager_.createColumn(TableManager.TABLE_TORRENT_PEERS, PID_COLUMN_ID);
		pidColumn_.initialize(TableColumn.ALIGN_LEAD, TableColumn.POSITION_LAST, 50);
		pidColumn_.setType(TableColumn.TYPE_TEXT_ONLY);
		pidColumn_.addCellRefreshListener(pidColumnRefreshListener_ = new PIDColumnRefreshListener());
		tableManager_.addColumn(pidColumn_);
	}

	private synchronized void removePIDColumn() {
		pidColumn_.removeCellRefreshListener(pidColumnRefreshListener_);
		/* FIXME: Is it possible to remove?? */
	}
	
	public synchronized void addView(UISWTInstance instance) {
		ViewListener view = new ViewListener(instance);
		instance.addView(UISWTInstance.VIEW_MAIN, VIEW_ID, view);
		openUIs_.add(instance);
	}

	public synchronized void removeView(UISWTInstance instance) {
		instance.removeViews(UISWTInstance.VIEW_MAIN, VIEW_ID);
		openUIs_.remove(instance);
	}

	public synchronized void cleanup() {
		removePIDColumn();
		uiManager_.removeUIListener(uiAttachListener_);
		while (!openUIs_.isEmpty())
			removeView(openUIs_.iterator().next());
	}

}
